/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of hierarchy information related functionality.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */


#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_HIERARCHY_INFO_C
#include <opensm/osm_sm.h>
#include <opensm/osm_unhealthy_ports.h>

#define hierarchy_info_level_mask(val, msb, lsb) \
	(((val) & (((1 << ((msb) - (lsb) + 1)) - 1) << lsb)) >> lsb)

#define SHORT_STR_LEN		10

static boolean_t osm_is_hierarchy_info_supported(IN osm_physp_t * p_physp)
{
	ib_net32_t capability_mask;
	osm_physp_t *p_physp0;

	if (!p_physp->p_node)
	       return FALSE;

	if (osm_node_get_type(p_physp->p_node) == IB_NODE_TYPE_SWITCH) {
		p_physp0 = osm_node_get_physp_ptr(p_physp->p_node, 0);
		capability_mask = p_physp0->port_info.capability_mask;
	} else {
		capability_mask = p_physp->port_info.capability_mask;
	}

	return (capability_mask & IB_PORT_CAP_HAS_HIER_INFO);
}

static void process_ndr_hierarchy_info(IN struct osm_sm *p_sm, IN osm_physp_t * p_physp)
{
	osm_hierarchy_info_t *p_hierarchy;
	char split_str[SHORT_STR_LEN] = {0};
	uint8_t node_type = osm_node_get_type(p_physp->p_node);
	uint8_t idx;
	uint8_t key;
	uint8_t max_levels;
	uint32_t level, value_23_8;
	int split, port, cage, slot_type, slot_serial_num, port_type;
	int bdf_bus, bdf_device, bdf_port;

	OSM_LOG_ENTER(p_sm->p_log);

	p_hierarchy = &p_physp->hierarchy_info;

	if (cl_ntoh64(p_hierarchy->block.template_guid) != OSM_HI_TEMPLATE_GUID_NDR)
		return;

	split = port = cage = slot_type = slot_serial_num = port_type =
	bdf_bus = bdf_device = bdf_port = -1;

	max_levels = ib_hierarchy_info_get_active_levels(&p_hierarchy->block);
	for (idx = 0; idx < max_levels; idx++) {
		level = cl_ntoh32(p_hierarchy->block.levels[idx]);
		/* bits lsb 8, msb 23 */
		value_23_8 = hierarchy_info_level_mask(level, 23, 8);
		/* bits lsb 24, msb 31 */
		key = hierarchy_info_level_mask(level, 31, 24);

		switch (key) {
			case HI_TEMPLATE_GUID_NDR_RECORD_SELECTOR_SPLIT:
				split = value_23_8;
				break;
			case HI_TEMPLATE_GUID_NDR_RECORD_SELECTOR_PORT:
				port = value_23_8;
				break;
			case HI_TEMPLATE_GUID_NDR_RECORD_SELECTOR_CAGE:
				cage = value_23_8;
				break;
			case HI_TEMPLATE_GUID_NDR_RECORD_SELECTOR_ASIC:
				p_hierarchy->asic = value_23_8;
				break;
			case HI_TEMPLATE_GUID_NDR_RECORD_SELECTOR_SLOT:
				/* bits lsb 0, msb 1 */
				slot_type = hierarchy_info_level_mask(level, 1, 0);
				slot_serial_num = value_23_8;
				break;
			case HI_TEMPLATE_GUID_NDR_RECORD_SELECTOR_TYPE:
				/* bits lsb 0, msb 0 */
				port_type = hierarchy_info_level_mask(level, 0, 0);
				break;
			case HI_TEMPLATE_GUID_NDR_RECORD_SELECTOR_BDF:
				/* bits lsb 16, msb 23 */
				bdf_bus = hierarchy_info_level_mask(level, 23, 16);
				/* bits lsb 11, msb 15 */
				bdf_device = hierarchy_info_level_mask(level, 15, 11);
				/* bits lsb 8, msb 10 */
				bdf_port = hierarchy_info_level_mask(level, 10, 8);
				break;
			default:
				break;
		}
	}
	/* switch and hca have to have 'cage' and 'port' for port label */
	if (cage == -1 || port == -1 ||
	    (node_type == IB_NODE_TYPE_SWITCH && p_hierarchy->asic == -1))
		return;

	/* reset previous value */
	memset(p_hierarchy->port_label, 0, sizeof(p_hierarchy->port_label));

	if (split != -1) {
		snprintf(split_str, SHORT_STR_LEN, "/%d", split);
	}

	if (node_type == IB_NODE_TYPE_SWITCH) {
		snprintf(p_hierarchy->port_label, OSM_PORT_LABEL_LEN, "%d/%d/%d%s",
			 p_hierarchy->asic, cage, port, split_str);
	} else if (node_type == IB_NODE_TYPE_CA) {
		/* slot type is physical */
		if (slot_type == 1 && slot_serial_num != -1) {
			snprintf(p_hierarchy->port_label, OSM_PORT_LABEL_LEN,
					"physical%d/%d/%d",
					slot_serial_num, cage, port);
		} else if (bdf_bus != -1 && bdf_device != -1 &&	bdf_port != -1) {
			snprintf(p_hierarchy->port_label, OSM_PORT_LABEL_LEN,
					"B%dD%dF%d/%d/%d%s",
					bdf_bus, bdf_device,
					bdf_port, cage, port,
					split_str);
		}
	}
	OSM_LOG_EXIT(p_sm->p_log);
}

#define OSM_HIERARCHY_INFO_PORT_TYPE_NUM       6
const char* g_xdr_port_type_strings[OSM_HIERARCHY_INFO_PORT_TYPE_NUM] = {"", "FNM1", "FNM", "sw", "HCA", "Access"};

static boolean_t check_xdr_template_guid_validity(IN struct osm_sm *p_sm,
						  IN OUT osm_physp_t * p_physp,
						  IN int port_type,
						  uint64_t template_guid)
{
	int aport = p_physp->hierarchy_info.aport;
	int plane_number = p_physp->hierarchy_info.plane_number;
	int num_of_planes = p_physp->hierarchy_info.num_of_planes;
	int asic = p_physp->hierarchy_info.asic;
	uint8_t num_of_ports = osm_node_get_num_physp(p_physp->p_node);
	uint8_t port_num = p_physp->port_num;
	osm_node_t *p_node = p_physp->p_node;
	boolean_t is_planarized = TRUE;

	/* this port has no planarization info */
	if (aport == -1) {
		p_physp->hierarchy_info.is_planarized = FALSE;
		return p_physp->hierarchy_info.is_planarized;
	}

	/* the aport is present, so the port is expected to be planarized */
	if (aport == 0 || aport >= num_of_ports) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR A007: "
			"Node 0x%" PRIx64 " port %u has invalid aport number %d "
			"(number of ports %u).\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)), port_num, aport, num_of_ports);
		is_planarized = FALSE;
	}

	if (num_of_planes <= 0 || num_of_planes > MAX_NUM_OF_PLANES) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR A008: "
			"Node 0x%" PRIx64 " aport %d port %u has invalid number of planes %d.\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)), aport,
			p_physp->port_num, num_of_planes);
		is_planarized = FALSE;
	}

	if (plane_number <= 0 || plane_number > MAX_NUM_OF_PLANES) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR A004: "
			"Node 0x%" PRIx64 " aport %d port %u has invalid plane number %d.\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)), aport,
			port_num, plane_number);
		is_planarized = FALSE;
	}

	if (template_guid == OSM_HI_TEMPLATE_GUID_4 &&
	    (asic <= 0 || asic > OSM_HI_MAX_NUM_OF_ASICS)) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR A009: "
			"Node 0x%" PRIx64 " aport %d port %u has invalid asic number %d.\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)), aport,
			port_num, asic);
		is_planarized = FALSE;
	}

	if (port_type < 0 || port_type >= OSM_HIERARCHY_INFO_PORT_TYPE_NUM) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR A00A: "
			"Node 0x%" PRIx64 " aport %d port %u has invalid port type value %d.\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)), aport,
			port_num, port_type);
	}

	p_physp->hierarchy_info.is_planarized = is_planarized;

	return p_physp->hierarchy_info.is_planarized;
}

static void process_hierarchy_info_template_guid_5(IN osm_sm_t *p_sm, IN osm_physp_t * p_physp,
						   IN uint64_t template_guid)
{
	osm_hierarchy_info_t *p_hierarchy;
	char cage_str[SHORT_STR_LEN] = {0}, ipil_str[SHORT_STR_LEN] = {0};
	char split_str[SHORT_STR_LEN] = {0}, bdf_str[2 * SHORT_STR_LEN] = {0};
	uint8_t node_type = osm_node_get_type(p_physp->p_node);
	uint8_t idx;
	uint8_t key;
	uint8_t max_levels;
	uint32_t level, value_23_8, value_15_8;
	int port_type, cage, ipil, split;
	int bdf_bus, bdf_device, bdf_port;

	OSM_LOG_ENTER(p_sm->p_log);

	p_hierarchy = &p_physp->hierarchy_info;

	if (cl_ntoh64(p_hierarchy->block.template_guid) != OSM_HI_TEMPLATE_GUID_5)
		return;

	port_type = cage = ipil = split = bdf_bus = bdf_device = bdf_port = -1;

	max_levels = ib_hierarchy_info_get_active_levels(&p_hierarchy->block);
	for (idx = 0; idx < max_levels; idx++) {
		level = cl_ntoh32(p_hierarchy->block.levels[idx]);
		/* bits lsb 8, msb 23 */
		value_23_8 = hierarchy_info_level_mask(level, 23, 8);
		value_15_8 = hierarchy_info_level_mask(level, 15, 8);
		/* bits lsb 24, msb 31 */
		key = hierarchy_info_level_mask(level, 31, 24);

		switch (key) {
			case HI_TEMPLATE_GUID_5_RECORD_SELECTOR_PORT_TYPE:
				port_type = value_15_8;
				break;
			case HI_TEMPLATE_GUID_5_RECORD_SELECTOR_CAGE:
				cage = value_23_8;
				break;
			case HI_TEMPLATE_GUID_5_RECORD_SELECTOR_IPIL:
				ipil = value_15_8;
				break;
			case HI_TEMPLATE_GUID_5_RECORD_SELECTOR_SPLIT:
				split = value_15_8;
				break;
			case HI_TEMPLATE_GUID_5_RECORD_SELECTOR_BDF:
				bdf_bus = hierarchy_info_level_mask(level, 23, 16);
				bdf_device = hierarchy_info_level_mask(level, 15, 11);
				bdf_port = hierarchy_info_level_mask(level, 10, 8);
				break;
			case HI_TEMPLATE_GUID_5_RECORD_SELECTOR_PLANE:
				p_hierarchy->plane_number = value_23_8;
				break;
			case HI_TEMPLATE_GUID_5_RECORD_SELECTOR_NUM_OF_PLANES:
				p_hierarchy->num_of_planes = value_23_8;
				break;
			case HI_TEMPLATE_GUID_5_RECORD_SELECTOR_APORT:
				p_hierarchy->aport = value_23_8;
				break;
			default:
				break;
		}
	}

	check_xdr_template_guid_validity(p_sm, p_physp, port_type, template_guid);

	/* switch and hca have to have 'port_type'. If none, use empty string */
	if (port_type < 0 || port_type >= OSM_HIERARCHY_INFO_PORT_TYPE_NUM)
		port_type = 0;

	if (cage != -1)
		snprintf(cage_str, sizeof(cage_str), "%d", cage);

	if (ipil != -1)
		snprintf(ipil_str, sizeof(ipil_str), "p%d", ipil);

	if (split != -1)
		snprintf(split_str, sizeof(split_str), "s%d", split);

	if (bdf_bus != -1 && bdf_device != -1 && bdf_port != -1)
		snprintf(bdf_str, sizeof(bdf_str), "ibB%dD%dF%d",
			 bdf_bus, bdf_device, bdf_port);

	/* reset previous value */
	memset(p_hierarchy->port_label, 0, sizeof(p_hierarchy->port_label));

	/* ibB{bus}D{device}F{function}{PortType}{$cage}p{IPIL}s{split} */
	if (node_type == IB_NODE_TYPE_CA) {
		snprintf(p_hierarchy->port_label, OSM_PORT_LABEL_LEN, "%s%s%s%s%s",
			 bdf_str, g_xdr_port_type_strings[port_type],
			 cage_str, ipil_str, split_str);
	}

	OSM_LOG_EXIT(p_sm->p_log);
}

static void process_hierarchy_info_template_guid_4(IN osm_sm_t *p_sm, IN osm_physp_t * p_physp,
						   IN uint64_t template_guid)
{
	osm_hierarchy_info_t *p_hierarchy;
	char asic_name_str[SHORT_STR_LEN] = {0}, ibport_str[SHORT_STR_LEN] = {0};
	char cage_str[SHORT_STR_LEN] = {0}, ipil_str[SHORT_STR_LEN] = {0};
	char split_str[SHORT_STR_LEN] = {0};
	uint8_t node_type = osm_node_get_type(p_physp->p_node);
	uint8_t idx;
	uint8_t key;
	uint8_t max_levels;
	uint32_t level, value_23_8, value_15_8;
	int port_type, asic_name, ib_port, cage, ipil, split;

	OSM_LOG_ENTER(p_sm->p_log);

	p_hierarchy = &p_physp->hierarchy_info;

	if (cl_ntoh64(p_hierarchy->block.template_guid) != OSM_HI_TEMPLATE_GUID_4)
		return;

	port_type = asic_name = ib_port = cage = ipil = split = -1;

	max_levels = ib_hierarchy_info_get_active_levels(&p_hierarchy->block);
	for (idx = 0; idx < max_levels; idx++) {
		level = cl_ntoh32(p_hierarchy->block.levels[idx]);
		/* bits lsb 8, msb 23 */
		value_23_8 = hierarchy_info_level_mask(level, 23, 8);
		value_15_8 = hierarchy_info_level_mask(level, 15, 8);
		/* bits lsb 24, msb 31 */
		key = hierarchy_info_level_mask(level, 31, 24);

		switch (key) {
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_PORT_TYPE:
				port_type = value_15_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_ASIC_NAME:
				asic_name = value_23_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_IB_PORT:
				ib_port = value_23_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_SWITCH_CAGE:
				cage = value_23_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_IPIL:
				ipil = value_15_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_SPLIT:
				split = value_15_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_ASIC:
				p_hierarchy->asic = value_15_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_TYPE:
				p_hierarchy->port_location_type =
					hierarchy_info_level_mask(level, 7, 0);
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_PLANE:
				p_hierarchy->plane_number = value_23_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_NUM_OF_PLANES:
				p_hierarchy->num_of_planes = value_23_8;
				break;
			case HI_TEMPLATE_GUID_4_RECORD_SELECTOR_APORT:
				p_hierarchy->aport = value_23_8;
				break;
			default:
				break;
		}
	}

	check_xdr_template_guid_validity(p_sm, p_physp, port_type, template_guid);

	/* switch and hca have to have 'port_type'. If none, use empty string */
	if (port_type < 0 || port_type >= OSM_HIERARCHY_INFO_PORT_TYPE_NUM)
		port_type = 0;

	if (asic_name >= 1 && asic_name <= 4)
		snprintf(asic_name_str, sizeof(asic_name_str), "A%d", asic_name);
	else if (asic_name == 256)
		snprintf(asic_name_str, sizeof(asic_name_str), "A");
	else if (asic_name == 257)
		snprintf(asic_name_str, sizeof(asic_name_str), "B");
	else if (asic_name != -1)
		snprintf(asic_name_str, sizeof(asic_name_str), "%d", asic_name);

	if (ib_port != -1)
		snprintf(ibport_str, sizeof(ibport_str), "P%d", ib_port);

	if (cage != -1)
		snprintf(cage_str, sizeof(cage_str), "%d", cage);

	if (ipil != -1)
		snprintf(ipil_str, sizeof(ipil_str), "p%d", ipil);

	if (split != -1)
		snprintf(split_str, sizeof(split_str), "s%d", split);


	/* reset previous value */
	memset(p_hierarchy->port_label, 0, sizeof(p_hierarchy->port_label));

	/* {PortType}{AsicName}P{Ibport}{SwitchCage}p{IPIL}sp{Split} */
	if (node_type == IB_NODE_TYPE_SWITCH) {
		snprintf(p_hierarchy->port_label, OSM_PORT_LABEL_LEN, "%s%s%s%s%s%s",
			 g_xdr_port_type_strings[port_type], asic_name_str, ibport_str,
			 cage_str, ipil_str, split_str);

	}

	OSM_LOG_EXIT(p_sm->p_log);
}

void osm_get_hierarchy_info(IN osm_sm_t * sm, IN osm_node_t * p_node,
			    IN osm_port_t * p_port,
			    IN osm_physp_t * p_physp,
			    IN uint8_t index)
{
	osm_madw_context_t context;
	osm_physp_t *p_physp0;
	ib_net64_t m_key;
	osm_madw_t *p_madw;
	uint8_t port_num = osm_physp_get_port_num(p_physp);
	uint32_t attr_mod = (index << IB_HIERARCHY_INFO_ATTR_MOD_INDEX_SHIFT); /* bits 8 - 15 */

	p_physp->hierarchy_info_updated =
		p_physp->hierarchy_info_updated &&
		!sm->p_subn->coming_out_of_standby &&
		!p_physp->need_update;

	if (p_physp->hierarchy_info_updated)
		return;

	/* check capability only for the first block, assuming that index 0 is always sent first */
	if (index == 0 && !osm_is_hierarchy_info_supported(p_physp))
		return;

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
		p_physp0 = osm_node_get_physp_ptr(p_node, 0);
		m_key = ib_port_info_get_m_key(&p_physp0->port_info);
		/* bits 0-7 */
		attr_mod |= (IB_HIERARCHY_INFO_ATTR_MOD_PORT_MASK & port_num);
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"SWITCH HI Node 0x%" PRIx64 " port number %d attr_mod %u\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)), port_num, attr_mod);
	} else {
		m_key = ib_port_info_get_m_key(&p_physp->port_info);
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"CA HI Node 0x%" PRIx64 " port number %d attr_mod %u\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)), port_num, attr_mod);
	}

	memset(&context, 0, sizeof(osm_madw_context_t));
	context.pi_context.node_guid = osm_node_get_node_guid(p_node);
	context.pi_context.port_guid = osm_physp_get_port_guid(p_physp);
	context.pi_context.set_method = FALSE;
	context.pi_context.light_sweep = FALSE;
	context.pi_context.active_transition = FALSE;
	context.pi_context.client_rereg = FALSE;
	context.pi_context.port_num = port_num;

	p_madw = osm_prepare_req_get(sm, osm_physp_get_dr_path_ptr(sm, p_physp),
			IB_MAD_ATTR_HIERARCHY_INFO,
			cl_hton32(attr_mod), /* query index 0 */
			FALSE, m_key, 0, CL_DISP_MSGID_NONE, &context);
	if (!p_madw)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A001: "
			"Failure initiating Hierarchical Info request\n");
	else
		cl_qlist_insert_tail(&p_port->mads_queue, &p_madw->list_item);
}

static boolean_t validate_hierarchy_info(osm_log_t *p_log, osm_node_t *p_node,
					 uint8_t port_num, ib_hierarchy_info_t *p_hi,
					 uint8_t hierarchy_index)
{
	if (p_hi->max_active_index > p_hi->max_supported_index) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR A00B: "
			"Port 0x%" PRIx64 " port num %u, hierarchy index %u, max active index (%u)"
			" is greater than max supported index (%u).\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)), port_num, hierarchy_index,
			p_hi->max_active_index, p_hi->max_supported_index);
		return FALSE;
	}
	return TRUE;
}

void osm_hierarchy_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_hierarchy_info_t *p_hi;
	ib_smp_t *p_smp;
	osm_pi_context_t *p_context;
	uint8_t port_num;
	osm_port_t *p_port;
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	uint8_t hierarchy_index;
	uint64_t template_guid;

	OSM_ASSERT(sm);

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_madw);

	p_smp = osm_madw_get_smp_ptr(p_madw);

	OSM_ASSERT(p_smp->attr_id == IB_MAD_ATTR_HIERARCHY_INFO);

	p_context = osm_madw_get_pi_context_ptr(p_madw);
	p_hi = ib_smp_get_payload_ptr(p_smp);

	port_num = p_context->port_num;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	p_port = osm_get_port_by_guid(sm->p_subn, p_context->port_guid);
	if (PF(!p_port)) {
		CL_PLOCK_RELEASE(sm->p_lock);
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A002: "
			"No port object for port with GUID 0x%" PRIx64
			"\n\t\t\t\tfor parent node GUID 0x%" PRIx64
			", TID 0x%" PRIx64 "\n",
			cl_ntoh64(p_context->port_guid),
			cl_ntoh64(p_context->node_guid), cl_ntoh64(p_smp->trans_id));
		goto Exit;
	}

	p_node = p_port->p_node;
	OSM_ASSERT(p_node);

	p_physp = osm_node_get_physp_ptr(p_node, port_num);

	OSM_ASSERT(p_physp);

	/* process only valid hierarchy info responses */
	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received. Send next mad.\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		goto Exit;
	}

	/* bits 15 - 8 */
	hierarchy_index = (cl_ntoh32(p_smp->attr_mod) & IB_HIERARCHY_INFO_ATTR_MOD_INDEX_MASK) >>
		IB_HIERARCHY_INFO_ATTR_MOD_INDEX_SHIFT;

	if (!validate_hierarchy_info(sm->p_log, p_node, port_num, p_hi, hierarchy_index))
		goto Exit;

	template_guid = cl_ntoh64(p_hi->template_guid);
	if (template_guid == OSM_HI_TEMPLATE_GUID_NDR) {
		p_physp->hierarchy_info.block = *p_hi;
		process_ndr_hierarchy_info(sm, p_physp);
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Node 0x%" PRIx64 " port number %d label %s\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)), port_num,
			p_physp->hierarchy_info.port_label);
	} else if (template_guid == OSM_HI_TEMPLATE_GUID_4) {
		p_physp->hierarchy_info.block = *p_hi;
		process_hierarchy_info_template_guid_4(sm, p_physp, template_guid);
	} else if (template_guid == OSM_HI_TEMPLATE_GUID_5) {
		p_physp->hierarchy_info.block = *p_hi;
		process_hierarchy_info_template_guid_5(sm, p_physp, template_guid);
	}

	if ((hierarchy_index + 1) <= p_hi->max_active_index) {
		osm_get_hierarchy_info(sm, p_node, p_port, p_physp, (hierarchy_index + 1));
	}

	/* if the block with the maximal index is received, all blocks were received */
	if (hierarchy_index == p_hi->max_active_index) {
		p_physp->hierarchy_info_updated = TRUE;
	}

Exit:
	/*
	 * Send next mad from the queue both on successful MAD and one with error status,
	 * in order not to disturb discovery.
	 */
	osm_port_send_next_mad(sm, p_port);

	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
}

osm_system_t *osm_system_new(IN const ib_node_info_t *p_ni)
{
	uint32_t size;
	osm_system_t *p_system;

	/*
	 * Create number of aports as the maximal possible number,
	 * which is equal to number of ports.
	 * The real number of aports will be most likely less than that number.
	 */
	size = p_ni->num_ports;

	p_system = malloc(sizeof(*p_system) + sizeof(osm_aport_t) * size);
	if (!p_system)
		return NULL;

	memset(p_system, 0, sizeof(*p_system) + sizeof(osm_aport_t) * size);

	p_system->system_guid = p_ni->sys_guid;
	p_system->aports_number = size;

	return p_system;
}

static void osm_build_aport(cl_map_item_t * const p_item, void *context)
{
	osm_node_t *p_node = (osm_node_t *)p_item;
	osm_sm_t *sm = (osm_sm_t *)context;
	uint8_t port_num;
	osm_physp_t *p_physp;
	osm_hierarchy_info_t *p_hierarchy;
	ib_node_info_t *node_info = &p_node->node_info;
	ib_net64_t sys_guid = node_info->sys_guid;
	osm_system_t *p_system;
	osm_aport_t *p_aport;
	uint8_t num_of_ports;

	num_of_ports = osm_node_get_num_physp(p_node);

	for (port_num = 1; port_num < num_of_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_node, port_num);

		if (!p_physp)
			continue;

		/* this port has no valid planarization info */
		if (!osm_hierarchy_is_planarized_physp(p_physp))
			continue;

		p_hierarchy = &p_physp->hierarchy_info;

		p_system = (osm_system_t *) cl_qmap_get(&sm->p_subn->system_guid_tbl, sys_guid);
		if (p_system == (osm_system_t *)cl_qmap_end(&sm->p_subn->system_guid_tbl)) {
			p_system = osm_system_new(&p_node->node_info);
			if (PF(p_system == NULL)) {
				OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A003: "
					"Unable to create new system object\n");
				return;
			}
			cl_qmap_insert(&sm->p_subn->system_guid_tbl, sys_guid, &p_system->map_item);
		}

		/* for precaution, all nodes of system are supposed to have the same number of ports */
		if (p_hierarchy->aport >= p_system->aports_number) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A006: "
				"Node 0x%" PRIx64 " port num %u's  hierarchy info with aport %u, "
				"that is greater than number of ports (%u) on nodes of system 0x%" PRIx64 ".\n",
				cl_ntoh64(osm_node_get_node_guid(p_node)), port_num, p_hierarchy->aport,
				p_system->aports_number, cl_ntoh64(sys_guid));
			continue;
		}

		p_aport = &p_system->aport_table[p_hierarchy->aport];

		if (p_aport->min_found_plane_number == 0)
			p_aport->min_found_plane_number = p_hierarchy->plane_number;

		/* check if aport has already physp there */
		if (p_aport->plane_ports[p_hierarchy->plane_number]) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A005: "
				"System 0x%" PRIx64 " node 0x%" PRIx64
				"aport %d port %u has the same plane %d as port %u.\n",
				cl_ntoh64(sys_guid), cl_ntoh64(osm_node_get_node_guid(p_node)),
				p_hierarchy->aport, p_physp->port_num, p_hierarchy->plane_number,
				p_aport->plane_ports[p_hierarchy->plane_number]->port_num);
		} else {
			p_aport->plane_ports[p_hierarchy->plane_number] = p_physp;

			if (p_hierarchy->aport > p_system->max_aport_available)
				p_system->max_aport_available = p_hierarchy->aport;
			
			if (p_hierarchy->plane_number >	p_aport->max_found_plane_number)
				p_aport->max_found_plane_number = p_hierarchy->plane_number;
		}
	}
}

void osm_system_delete(IN OUT osm_system_t ** p_system)
{
	OSM_ASSERT(p_system && *p_system);
	free(*p_system);
	*p_system = NULL;
}

extern void osm_subn_destroy_system_table(IN osm_subn_t * p_subn);

void osm_build_aports(osm_sm_t * sm)
{
	uint32_t systems_num = 0;

	OSM_LOG_ENTER(sm->p_log);

	/* clear previous db */
	osm_subn_destroy_system_table(sm->p_subn);

	cl_qmap_apply_func(&sm->p_subn->node_guid_tbl, osm_build_aport, sm);

	systems_num = cl_qmap_count(&sm->p_subn->system_guid_tbl);
	if (systems_num) {
		sm->p_subn->is_planarized = TRUE;
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Planarized systems (%u) found.\n", systems_num);
	}

	OSM_LOG_EXIT(sm->p_log);
}

static void osm_analyze_aports(cl_map_item_t * const p_item, void *context)
{
	osm_system_t *p_system = (osm_system_t *)p_item;
	osm_sm_t *sm = (osm_sm_t *)context;
	osm_aport_t *p_aport;
	osm_physp_t *p_physp;
	osm_node_t *p_node;
	uint8_t aport_num;
	uint8_t plane_port;
	uint8_t port_state = 0, phys_port_state = 0;
	uint8_t first_port_link_speed_active, port_link_speed_active;
	uint8_t link_width_active = 0, mtu_cap = 0, vl_cap = 0;
	boolean_t first_port_found = FALSE;
	osm_unhealthy_port_t *p_uhp;
	boolean_t is_heldback_switch;
	boolean_t has_fdr10 = (1 == sm->p_subn->opt.fdr10) ? TRUE : FALSE;

	OSM_LOG_ENTER(sm->p_log);

	for (aport_num = 1; aport_num <= p_system->max_aport_available; aport_num++) {
		p_aport = &p_system->aport_table[aport_num];
		/* reset for new aport */
		port_state = phys_port_state = first_port_link_speed_active = port_link_speed_active =
		link_width_active = mtu_cap = vl_cap = 0;

		first_port_found = FALSE;

		for (plane_port = 1; plane_port <= p_aport->max_found_plane_number; plane_port++) {
			p_physp = p_aport->plane_ports[plane_port];

			if (!p_physp) {
				continue;
			}

			p_node = p_physp->p_node;

			/*
			 * look at parameters:
			 * Relevant attribute values are:
			 * active speed, active width, MTU cap and VL cap.
			 */
			phys_port_state = ib_port_info_get_port_state(&p_physp->port_info);

			/* first found port, lets remember its values for comparison */
			if (!first_port_found) {
				port_state = phys_port_state;
				first_port_link_speed_active =
					osm_physp_get_highest_active_speed(p_physp, has_fdr10);
				link_width_active = p_physp->port_info.link_width_active;
				mtu_cap = ib_port_info_get_mtu_cap(&p_physp->port_info);
				vl_cap = ib_port_info_get_vl_cap(&p_physp->port_info);
				first_port_found = TRUE;
				continue;
			}

			port_link_speed_active =
				osm_physp_get_highest_active_speed(p_physp, has_fdr10);

			/*
			 * Check that the port that has different link state is not just
			 * reconnected port that is in init state and is about to move to active after routing.
			 */
			if ((phys_port_state != port_state &&
			        !(phys_port_state == IB_LINK_INIT && p_physp->need_update)) ||
			    port_link_speed_active != first_port_link_speed_active ||
			    p_physp->port_info.link_width_active != link_width_active ||
			    ib_port_info_get_mtu_cap(&p_physp->port_info) != mtu_cap ||
			    ib_port_info_get_vl_cap(&p_physp->port_info) != vl_cap) {

				p_aport->no_route = TRUE;
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Found asymmetric aport %u on System GUID 0x%" PRIx64
					" node GUID 0x%" PRIx64
					" port number %u (need update %u), plane %u, state: %u port's %u, "
					"speed %u port's %u, "
					"width: %u port's %u, mtu_cap %u port's %u, vl_cap %u port's %u\n",
					aport_num, cl_ntoh64(p_node->node_info.sys_guid),
					cl_ntoh64(osm_node_get_node_guid(p_node)),
					p_physp->port_num, p_physp->need_update, plane_port,
					port_state, phys_port_state,
					first_port_link_speed_active, port_link_speed_active,
					link_width_active, p_physp->port_info.link_width_active,
					mtu_cap, ib_port_info_get_mtu_cap(&p_physp->port_info),
					vl_cap, ib_port_info_get_vl_cap(&p_physp->port_info));
				break;
			}

			p_uhp = osm_hm_get_unhealthy_port_by_guid(sm,
								  cl_hton64(osm_node_get_node_guid(p_node)),
								  p_physp->port_num);
			if (p_uhp && (p_uhp->isolate || p_uhp->no_discover)) {
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Found unhealthy aport %u on System GUID 0x%" PRIx64
					" node GUID 0x%" PRIx64	" port number %u, "
					"isolate %u no_discover %u\n",
					aport_num, cl_ntoh64(p_node->node_info.sys_guid),
					cl_ntoh64(osm_node_get_node_guid(p_node)),
					p_physp->port_num,
					p_uhp->isolate, p_uhp->no_discover);
				p_aport->no_route = TRUE;
				break;
			}

			is_heldback_switch = lookup_held_back_switch(sm->p_subn, p_node);
			if (p_physp->port_reset_required || is_heldback_switch) {
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Found 'reset required' aport %u on System GUID 0x%" PRIx64
					" node GUID 0x%" PRIx64	" with plane port %u "
					"port number %u, reset_required %u is_heldback_switch %u\n",
					aport_num, cl_ntoh64(p_node->node_info.sys_guid),
					cl_ntoh64(osm_node_get_node_guid(p_node)),
					p_physp->hierarchy_info.plane_number,
					p_physp->port_num,
					p_physp->port_reset_required, is_heldback_switch);
				p_aport->no_route = TRUE;
				break;
			}
		}
	}

	OSM_LOG_EXIT(sm->p_log);
}

void osm_analyze_systems_aports(IN osm_sm_t * sm)
{
	cl_qmap_apply_func(&sm->p_subn->system_guid_tbl, osm_analyze_aports, sm);
}

boolean_t osm_hierarchy_is_planarized_physp(IN struct osm_physp *p_physp)
{
	return osm_physp_is_valid(p_physp) && p_physp->hierarchy_info.is_planarized;
}
