/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Main entry point into routing subsystem.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTING_C
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_routing.h>
#include <opensm/osm_advanced_routing.h>

typedef int (*f_re_setup_t) (struct osm_routing_engine *, osm_opensm_t *);

struct routing_engine_module {
	const char *name;
	f_re_setup_t setup;
};

extern int osm_ucast_minhop_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_updn_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_dnup_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_file_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_ftree_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_pqft_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_lash_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_dor_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_torus2QoS_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_sssp_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_dfsssp_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_chain_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_dfp_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_kdor_hc_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_kdor_ghc_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_dfp2_setup(struct osm_routing_engine *, osm_opensm_t *);

extern int osm_ucast_armgr_lag_setup(osm_adv_routing_engine_t *, osm_subn_t *);
extern int osm_ucast_armgr_tree_setup(osm_adv_routing_engine_t *, osm_subn_t *);
extern int osm_ucast_armgr_minhop_setup(osm_adv_routing_engine_t *, osm_subn_t *);

extern int osm_ucast_ar_updn_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_ar_ftree_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_ar_torus_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_ar_dor_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_ar_minhop_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_auto_engine_setup(struct osm_routing_engine *, osm_opensm_t *);

/*
 * Routing engine modules.
 */
static const struct routing_engine_module routing_modules[] = {
	[ OSM_ROUTING_ENGINE_TYPE_NONE ]        = {  "none",       NULL                         },
	[ OSM_ROUTING_ENGINE_TYPE_MINHOP ]      = {  "minhop",     osm_ucast_minhop_setup       },
	[ OSM_ROUTING_ENGINE_TYPE_UPDN ]        = {  "updn",       osm_ucast_updn_setup         },
	[ OSM_ROUTING_ENGINE_TYPE_DNUP ]        = {  "dnup",       osm_ucast_dnup_setup         },
	[ OSM_ROUTING_ENGINE_TYPE_FILE ]        = {  "file",       osm_ucast_file_setup         },
	[ OSM_ROUTING_ENGINE_TYPE_FTREE ]       = {  "ftree",      osm_ucast_ftree_setup        },
	[ OSM_ROUTING_ENGINE_TYPE_PQFT ]        = {  "pqft",       osm_ucast_pqft_setup         },
	[ OSM_ROUTING_ENGINE_TYPE_LASH ]        = {  "lash",       osm_ucast_lash_setup         },
	[ OSM_ROUTING_ENGINE_TYPE_DOR ]         = {  "dor",        osm_ucast_dor_setup          },
	[ OSM_ROUTING_ENGINE_TYPE_TORUS_2QOS ]  = {  "torus-2QoS", osm_ucast_torus2QoS_setup    },
	[ OSM_ROUTING_ENGINE_TYPE_DFSSSP ]      = {  "dfsssp",     osm_ucast_dfsssp_setup       },
	[ OSM_ROUTING_ENGINE_TYPE_SSSP ]        = {  "sssp",       osm_ucast_sssp_setup         },
	[ OSM_ROUTING_ENGINE_TYPE_CHAIN ]       = {  "chain",      osm_ucast_chain_setup        },
	[ OSM_ROUTING_ENGINE_TYPE_DFP ]         = {  "dfp",        osm_ucast_dfp_setup          },
	[ OSM_ROUTING_ENGINE_TYPE_DFP2 ]        = {  "dfp2",       osm_ucast_dfp2_setup         },
	[ OSM_ROUTING_ENGINE_TYPE_AR_DOR ]      = {  "ar_dor",     osm_ucast_ar_dor_setup       },
	[ OSM_ROUTING_ENGINE_TYPE_AR_UPDN ]     = {  "ar_updn",    osm_ucast_ar_updn_setup      },
	[ OSM_ROUTING_ENGINE_TYPE_AR_FTREE ]    = {  "ar_ftree",   osm_ucast_ar_ftree_setup     },
	[ OSM_ROUTING_ENGINE_TYPE_AR_TORUS ]    = {  "ar_torus",   osm_ucast_ar_torus_setup     },
	[ OSM_ROUTING_ENGINE_TYPE_KDOR_HC ]     = {  "kdor-hc",    osm_ucast_kdor_hc_setup      },
	[ OSM_ROUTING_ENGINE_TYPE_KDOR_GHC ]    = {  "kdor-ghc",   osm_ucast_kdor_ghc_setup     },
	[ OSM_ROUTING_ENGINE_TYPE_AR_MINHOP ]   = {  "ar_minhop",  osm_ucast_ar_minhop_setup    },
	[ OSM_ROUTING_ENGINE_TYPE_AUTO ]        = {  "auto",       osm_ucast_auto_engine_setup  },
};

const char *osm_routing_engine_type_str(IN osm_routing_engine_type_t type)
{
	if ((0 <= type) && (type < OSM_ROUTING_ENGINE_TYPE_END)) {
		return routing_modules[type].name;
	}

	return "unknown";
}

osm_routing_engine_type_t osm_routing_engine_type(IN const char *name)
{
	int i;

	/*
	 * For legacy reasons, consider a NULL pointer and the string
	 * "null" as the minhop routing engine.
	 */
	if ((name == NULL) || (strcasecmp(name, "null") == 0)) {
		name = "minhop";
	}

	for (i = 0; i < OSM_ROUTING_ENGINE_TYPE_END; i++) {
		if (strcasecmp(name, routing_modules[i].name) == 0) {
			return i;
		}
	}

	return -1;
}

f_re_setup_t osm_routing_engine_setup(IN const char *name)
{
	int type = osm_routing_engine_type(name);
	return (type >= 0) ? routing_modules[type].setup : NULL;
}

const struct routing_engine_module *osm_routing_engine_module(IN const char *name)
{
	int type = osm_routing_engine_type(name);
	return (type >= 0) ? &routing_modules[type] : NULL;
}

/*
 * Advanced routing engine modules.
 */
typedef int(*f_adv_re_setup_t)(osm_adv_routing_engine_t *, osm_subn_t *);

struct adv_routing_engine_module {
        const char *name;
        f_adv_re_setup_t setup;
};


static const struct adv_routing_engine_module adv_routing_modules[] = {
	[ OSM_ADV_ROUTING_TYPE_NONE ]           = {  "none",       NULL                          },
	[ OSM_ADV_ROUTING_TYPE_AR_LAG ]         = {  "ar_lag",     osm_ucast_armgr_lag_setup     },
	[ OSM_ADV_ROUTING_TYPE_AR_TREE ]        = {  "ar_tree",    osm_ucast_armgr_tree_setup    },
	[ OSM_ADV_ROUTING_TYPE_AR_MINHOP ]      = {  "ar_minhop",  osm_ucast_armgr_minhop_setup  },
};

const char *get_adv_routing_name(osm_adv_routing_type_enum type)
{
	if ((type < 0) || (type >= OSM_ADV_ROUTING_TYPE_END)) {
		type = 0;
	}

	return adv_routing_modules[type].name;
}

osm_adv_routing_type_enum get_adv_routing_type(const char *name)
{
	int i;

	if ((name == NULL) || (strcasecmp(name, "null") == 0)) {
		name = "none";
	}

	for (i = 0; i < OSM_ADV_ROUTING_TYPE_END; i++) {
		if (strcasecmp(name, adv_routing_modules[i].name) == 0) {
			return i;
		}
	}

	return OSM_ADV_ROUTING_TYPE_NONE;
}

f_adv_re_setup_t get_adv_routing_setup(const char *name)
{
	int type = get_adv_routing_type(name);
	return adv_routing_modules[type].setup;
}

static void append_routing_engine(osm_subn_t *subn, struct osm_routing_engine *routing_engine)
{
	struct osm_routing_engine *r;

	routing_engine->next = NULL;

	if (!subn->routing_engine_list) {
		subn->routing_engine_list = routing_engine;
		return;
	}

	r = subn->routing_engine_list;
	while (r->next)
		r = r->next;

	r->next = routing_engine;
}

/* default callback function for retrieving routing engine name */
static const char *routing_engine_get_name(struct osm_routing_engine *re)
{
	return re->name;
}

/* default callback function for retrieving routing engine attributes */
static struct osm_routing_engine_attributes routing_engine_get_attributes(struct osm_routing_engine *re)
{
	return re->attributes;
}

struct osm_routing_engine *setup_routing_engine(osm_subn_t *subn, const char *name)
{
	struct osm_routing_engine *re;
	const struct routing_engine_module *p_module;  

	if (!strcmp(name, "no_fallback")) {
		subn->no_fallback_routing_engine = TRUE;
		return NULL;
	}

	/*
	 * Verify that the routing engine name is known.
	 */
	if (p_module = osm_routing_engine_module(name), p_module == NULL) {
		OSM_LOG(&subn->p_osm->log, OSM_LOG_ERROR, "cannot find or setup routing engine \'%s\'\n", name);
		return NULL;
	}

	/*
	 * Allocate the routing engine.
	 */
	if ((re = calloc(1, sizeof(struct osm_routing_engine))) == NULL) {
		OSM_LOG(&subn->p_osm->log, OSM_LOG_VERBOSE, "memory allocation failed\n");
		return NULL;
	}

	re->get_name = routing_engine_get_name;
	re->get_attributes = routing_engine_get_attributes;
	re->name = p_module->name;
	re->type = osm_routing_engine_type(name);

	/*
	 * Setup the routing engine.
	 */
	if ((p_module->setup != NULL) && (p_module->setup(re, subn->p_osm) != 0)) {
		OSM_LOG(&subn->p_osm->log, OSM_LOG_VERBOSE, "setup of routing engine \'%s\' failed\n", name);
		free(re);
		return NULL;
	}

	if (!subn->is_auto_engine_used && !strcmp(re->name, "auto")) {
		subn->is_auto_engine_used = TRUE;
	}

	OSM_LOG(&subn->p_osm->log, OSM_LOG_DEBUG, "\'%s\' routing engine set up\n", re->name);
	return re;
}

void osm_setup_routing_engines(osm_subn_t *subn, const char *engine_names)
{
	char *name, *str, *p;
	struct osm_routing_engine *re;

	subn->is_auto_engine_used = FALSE;

	if (engine_names && *engine_names) {
		str = strdup(engine_names);
		name = strtok_r(str, ", \t\n", &p);
		while (name && *name) {
			re = setup_routing_engine(subn, name);
			if (re) {
				append_routing_engine(subn, re);
				/* update is_auto_engine_used indicator */
				if (re->type == OSM_ROUTING_ENGINE_TYPE_AUTO)
					subn->is_auto_engine_used = TRUE;
			}
			name = strtok_r(NULL, ", \t\n", &p);
		}
		free(str);
	}
	if (!subn->default_routing_engine) {
		subn->default_routing_engine = setup_routing_engine(subn, "minhop");
	}
}

void osm_destroy_default_routing_engine(osm_subn_t *subn)
{
	struct osm_routing_engine *r;

	r = subn->default_routing_engine;
	if (r) {
		if (r->destroy && r->context)
			r->destroy(r->context);
		free(r);
		subn->default_routing_engine = NULL;
	}
}

void osm_destroy_routing_engines(osm_subn_t *subn)
{
	struct osm_routing_engine *r, *next;

	next = subn->routing_engine_list;
	while (next) {
		r = next;
		next = r->next;
		if (r != subn->default_routing_engine) {
			if (r->destroy)
				r->destroy(r->context);
			free(r);
		} else /*do not free default_routing_engine */
			r->next = NULL;
	}
	subn->routing_engine_list = NULL;
}

static void update_routing_engine(struct osm_routing_engine *cur, struct osm_routing_engine *last)
{
	struct osm_routing_engine *next = cur->next;

	if (!last)
		return; /* no last routing engine */

	/*
	 * since function overrides 'cur', we need to make sure 'cur' context is
	 * released.
	 */
	if (cur->context && cur->destroy)
		cur->destroy(cur->context);

	memcpy(cur, last, sizeof(*cur));
	/* restore next */
	cur->next = next;
}

void osm_update_routing_engines(osm_subn_t *subn, const char *engine_names)
{
	struct osm_routing_engine *r, *l;

	/* find used routing engine and save as last */
	l = r = subn->routing_engine_list;

	if (!r)
		goto config_engines;

	/* skip if routing_engine_used not set */
	if (subn->routing_engine_used == OSM_ROUTING_ENGINE_TYPE_NONE)
		goto destroy_engines;

	if (subn->routing_engine_used->type == osm_routing_engine_type(r->name)) {
		subn->last_routing_engine = r;
		subn->routing_engine_list = r->next;
	} else while ((r = r->next)) {
		if (subn->routing_engine_used->type ==
		    osm_routing_engine_type(r->name)) {
			subn->last_routing_engine = r;
			l->next = r->next;
			break;
		}
		l = r;
	}

destroy_engines:
	/* cleanup prev routing engine list and replace with current list */
	osm_destroy_routing_engines(subn);

config_engines:
	osm_setup_routing_engines(subn, engine_names);

	if (subn->routing_engine_used == OSM_ROUTING_ENGINE_TYPE_NONE)
		return;

	if (subn->routing_engine_used == subn->default_routing_engine)
		return;

	/* check if last routing engine exist in new list and update callbacks */
	r = subn->routing_engine_list;
	while (r) {
		if (subn->routing_engine_used->type ==
		    osm_routing_engine_type(r->name)) {
			update_routing_engine(r, subn->last_routing_engine);
			subn->last_routing_engine = NULL;
			free(subn->routing_engine_used);
			subn->routing_engine_used = r;
			break;
		}
		r = r->next;
	}
}

void osm_destroy_advanced_routing_engine(osm_subn_t *p_subn)
{
	if (p_subn->adv_routing_engine) {
		p_subn->adv_routing_engine->destroy(p_subn->adv_routing_engine->context);
		free(p_subn->adv_routing_engine);
		p_subn->adv_routing_engine = NULL;
	}
}

void osm_create_advanced_routing_engine(osm_subn_t *p_subn, const char *name)
{
	osm_adv_routing_engine_t *p_adv_re = NULL;
	f_adv_re_setup_t setup_function;

	/*
	 * Get the setup function.
	 */
	setup_function = get_adv_routing_setup(name);
	if (setup_function == NULL) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7535: cannot find advanced routing engine \'%s\'\n", name);
		return;
	}

	/*
	 * Construct the new advanced routing engine.
	 */
	if ((p_adv_re = (osm_adv_routing_engine_t *)calloc(1, sizeof(*p_adv_re))) == NULL) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR 7533: memory allocation failed\n");
		return;
	}

	/*
	 * Call the setup function.
	 */
	p_adv_re->name = name;
	if (setup_function(p_adv_re, p_subn)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7534: failed to initialize advanced routing engine module \'%s\'\n", name);
		free(p_adv_re);
		return;
	}

	p_subn->adv_routing_engine = p_adv_re;
}

void osm_setup_advanced_routing_engine(osm_subn_t *p_subn, const char *engine_name)
{
	/* destroy current advanced routing engine module */
	osm_destroy_advanced_routing_engine(p_subn);

	if (!strcasecmp(engine_name, "none") || !strcasecmp(engine_name, "auto")) {
		return;
	}

	osm_create_advanced_routing_engine(p_subn, engine_name);
}

